/* Compare inter arrival times of serveral pcap traces. Used to
   determine the "quality" of replaying tools like tcpreplay.
	Each given file is compared against each other.
	
	contact: <gregor@majordomus.org>
	
Copyright (c) 2005 Gregor Maier.  All rights reserved.


Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
3. Neither the names of the copyright owners nor the names of its
   contributors may be used to endorse or promote products derived from
   this software without specific prior written permission.
4. All advertising materials mentioning features or use of this software
   must display the following acknowledgement:
       This product includes software developed by Anzen Computing, Inc.

THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
	*/


#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <math.h>
#include <pcap.h>

#define MAXFILES 7
#define NUMBUCKETS 6
static const double bucket_bounds[] = {1e-6, 5e-6, 10e-6, 20e-6, 50e-6};

#define tv_to_double(tv) ((tv).tv_sec + 1e-6*(tv).tv_usec) 

struct result {
	int iat[NUMBUCKETS];
	int off[NUMBUCKETS];
	double sq_err;
	double avg_err;
};

int main (int argc, char* argv[]) 
{
	int i, j, k;
	char optstring[] = "hm:";
	int opt;
	int filecnt;  /* number of pcap files to compare */
	int combinations; /* number of possible combinations of input files */
	int curcomb;
	struct result *res;
	char *filenames[MAXFILES];
	pcap_t *pd[MAXFILES];
	char errbuf[PCAP_ERRBUF_SIZE+1];
	int wet;  /* set to true if at least one pcap file is not yet try */
	struct pcap_pkthdr hdrs[MAXFILES];
	double firstpkt[MAXFILES], lastpkt[MAXFILES], curpkt[MAXFILES];
	unsigned numpackets, maxpackets=-1;
	double iat1, off1, iat2, off2;
	int is_first=1;
	
	
	while ((opt=getopt(argc, argv, optstring))!=-1) {
		switch (opt) {
			case 'h':
				printf("iat_cmp - Read a pcap dump files and compare inter arrival times\n\n");
				printf("Usage: iat_cmp [-m maxpkts] [-h] files\n");
				printf("    -h             Print this message.\n");
				printf("    -m maxpkts     Read no more then maxpkts from the input files.\n");
				exit(0);
				break;
			case 'm':
				maxpackets = strtoul(optarg, NULL, 10);
				break;
			default:
				fprintf(stderr, "%s: getopt returned an unknown character code %x\n", argv[0], opt);
				break;
		}
	}
	filecnt = argc-optind;

	if(filecnt > MAXFILES) {
		fprintf(stderr, "Warning: more files (%d) given than allow (%d). Ignoring last %d files",
			filecnt, MAXFILES, MAXFILES-filecnt);
		filecnt=MAXFILES;
	}

	/* allocate the structs for the results and init them */
	combinations = (filecnt*(filecnt-1))/2;
	res = calloc(combinations, sizeof(struct result));
	for (i=0; i<combinations; i++) {
		for (j=0; j<NUMBUCKETS; j++) { 
			res[i].iat[j] = 0;
			res[i].off[j] = 0;
		}
		res[i].sq_err = 0.0;
		res[i].avg_err = 0.0;
	}
		

	/* Copy filenames and open pcap handles for all given files */
	for (i=0; i<filecnt; i++) {
		filenames[i] = strdup(argv[optind+i]);
		pd[i] = pcap_open_offline(filenames[i], errbuf);
		if (pd[i]==NULL) {
			fprintf(stderr, "%s: pcap_open_offline(%s): %s\n", argv[0], filenames[i], errbuf);
			exit(1);
		}
	}


	wet=1;
	numpackets=0;
	while(wet && numpackets<maxpackets) {
		wet = 0;
		/* Read the next packet from each source */
		for (i=0; i<filecnt; i++)	{
			if (pcap_next(pd[i], &hdrs[i])!=NULL) { 
				wet = 1;
				if (is_first) 
					firstpkt[i] = lastpkt[i] = tv_to_double(hdrs[i].ts);
				else {
					lastpkt[i] = curpkt[i];
					curpkt[i] = tv_to_double(hdrs[i].ts);
				}
			}
			else  /* gone try */
				curpkt[i] = -1e6;
		}
		if (!is_first) {
			curcomb = 0;
			for (i=0; i<filecnt; i++) {
				for (j=i+1; j<filecnt; j++) {
					iat1 = curpkt[i] - lastpkt[i];
					off1 = curpkt[i] - firstpkt[i];
					iat2 = curpkt[j] - lastpkt[j];
					off2 = curpkt[j] - firstpkt[j];
					for (k=0; k<NUMBUCKETS-1; k++) {
						if (fabs(iat1-iat2) < bucket_bounds[k]) {
							res[curcomb].iat[k]++;
							break;
						}
					}
					if (fabs(iat1-iat2) >= bucket_bounds[NUMBUCKETS-2])
						res[curcomb].iat[NUMBUCKETS-1]++;
					
					for (k=0; k<NUMBUCKETS-1; k++) {
						if (fabs(off1-off2) < bucket_bounds[k]) {
							res[curcomb].off[k]++;
							break;
						}
					}
					if (fabs(off1-off2) >= bucket_bounds[NUMBUCKETS-2])
						res[curcomb].off[NUMBUCKETS-1]++;

					res[curcomb].sq_err += (iat1-iat2)*(iat1-iat2);
					/*res[curcomb].avg_err += fabs(off1-off2);*/
					res[curcomb].avg_err += fabs(iat1-iat2);

					curcomb++;
				}
			}
		}
		is_first = 0;
		numpackets++;
	} /* end while */


	/*  Output the results */
	curcomb = 0;
	for (i=0; i<filecnt; i++) {
		for (j=i+1; j<filecnt; j++) {
			printf("Comparing: %s VS %s\n", filenames[i], filenames[j]);
			printf("    No of pkts:  %d\n", numpackets);
			printf("    Sqaure err:  %le --- Square err/pkt: %le\n", res[curcomb].sq_err, 
					res[curcomb].sq_err/numpackets); 
			printf("    Avg. err: %le\n", res[curcomb].avg_err/numpackets);
			for (k=0; k<NUMBUCKETS-1; k++) {
				printf("    IAT %le bucket: %d   (%5.2lf%%)\n", bucket_bounds[k], res[curcomb].iat[k],
					100.*res[curcomb].iat[k]/((double)numpackets));
				printf("    OFF %le bucket: %d\n", bucket_bounds[k], res[curcomb].off[k]);
			}
			printf("    IAT rem. bucket: %d\n", res[curcomb].iat[NUMBUCKETS-1]);
			printf("    OFF rem. bucket: %d\n", res[curcomb].off[NUMBUCKETS-1]);
			curcomb++;
			printf("=======================================================================\n");
		}
	}
			
	exit(0);
}
